#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>

#define MAX_STR 255
#define MAX_CASES 20

int CompilarPrograma(char* programa)
{
    int status;
    pid_t pid;
    
    if ((pid = fork()) == -1)
        exit(1);

    if (pid) { // processo pai
        // output do lado do processo principal, para evitar problemas de ordem das linhas de output
        printf("\nProcesso PID = %d PPID = %d [compilacao].", (int)pid, (int)getpid());
        fflush(stdout);
        wait(&status);
        if (WIFEXITED(status) && WEXITSTATUS(status) == 0) 
            // compilation ok
            return 1;
    }
    else // processo filho
        execlp("gcc", "gcc", "-Wall", "-o", "tst", programa, NULL);
    return 0;
}

void LimparFimDeLinha(char* str) {
    while(strlen(str)>0 && strchr("\n\r", str[strlen(str) - 1])!=NULL)
        str[strlen(str) - 1] = 0; 
}

int LerCasos(char* casos, 
    char input[MAX_CASES][MAX_STR], 
    char output[MAX_CASES][MAX_STR],
    int *ncasos) 
{
    char str[MAX_STR];
    FILE* f;
    *ncasos = 0;
    if ((f = fopen(casos, "rt")) != NULL) {
        while (!feof(f)) {
            if (fgets(str, MAX_STR, f) != NULL) {
                strcpy(input[*ncasos], str);
                if (fgets(str, MAX_STR, f) != NULL) {
                    LimparFimDeLinha(str);
                    strcpy(output[(*ncasos)++], str);
                }
            }
        }
        fclose(f);
    } 
    return *ncasos>0;
}

void ExecutarCaso(int i, char* arg)
{
    char *argv[10];
    char* pt;
    char str[MAX_STR];
    int j;
    pid_t pid;
    //FILE* f;

    if ((pid = fork()) == -1)
        exit(1);

    if (pid==0) { // processo filho
        // preparar os argumentos
        argv[0] = "tst";
        j = 1;
        pt = strtok(arg, " ");
        while (pt != NULL) {
            argv[j++] = pt;
            pt = strtok(NULL, " ");
        }
        argv[j] = NULL;

        // redirecionar o output
        sprintf(str, "out%d.txt", i);
        stdout = freopen(str, "wt", stdout);
        //if ((f = fopen(str, "wt")) != NULL) {
        //    dup2(fileno(f), STDOUT_FILENO);
        //    fclose(f);
        //}

        execvp("./tst", argv);
    }

    // processo pai sai de imediato, lan�ando outro caso de teste de seguida
    // colocar o output do lado do processo principal, para evitar problemas de ordem das linhas de output
    printf("\nProcesso PID = %d PPID = %d [execucao caso %d].", (int)pid, (int)getpid(), i);
    fflush(stdout);
}

void VerificarCaso(int i, char* output)
{
    char str[MAX_STR];
    FILE* f;
    sprintf(str, "out%d.txt", i);
    if ((f = fopen(str, "rt")) != NULL) {
        strcpy(str, "");
        fgets(str, MAX_STR, f);
        LimparFimDeLinha(str);
        if (strstr(str, output) == NULL)
            printf("\nErro caso %d: esperado '%s' obtido '%s'.", i, output, str);
    }
    else
        printf("\nErro caso %d nao executado.", i);
}


int main(int argc, char **argv)
{
    char input[MAX_CASES][MAX_STR];
    char output[MAX_CASES][MAX_STR];
    int ncasos, i, status;

    if (argc != 3) {
        printf("\nUtilizacao: teste <programa.c> <casos.txt>");
        exit(0);
    }

    printf("\nProcesso PID = %d PPID = %d [principal].", (int)getpid(), (int)getppid());

    // 1. compilar o programa
    if (CompilarPrograma(argv[1])) {
        // 2. ler o ficheiro casos.txt
        if (LerCasos(argv[2], input, output, &ncasos)) {
            // 3. executar os casos de teste, um processo por cada caso de teste
            for (i = 0; i < ncasos; i++)
                ExecutarCaso(i, input[i]);

            // 4. aguardar que todos os processos terminem
            for (i = 0; i < ncasos; i++)
                wait(&status);

            // 5. processar resultados
            for (i = 0; i < ncasos; i++)
                VerificarCaso(i, output[i]);

            printf("\nCasos de teste executados: %d.\n", ncasos);
        }
        else
            printf("\nErro ao ler os casos de teste.\n");
    }
    else
        printf("\nErro na compilacao.\n");

    return 0;
}